# pylint: disable=unused-argument,multiple-statements,no-self-use,super-init-not-called
from __future__ import annotations

import sys
from typing import Pattern, Any, List, NamedTuple

if sys.version_info >= (3, 8):
    from typing import TypedDict, Literal
else:
    from typing_extensions import TypedDict, Literal

ATTR_NAME_REGEX: Pattern[str]

ConditionBaseExpressionType = TypedDict(
    "ConditionBaseExpressionType", {"format": str, "operator": str, "values": List[Any]}
)

class ConditionBase:
    expression_format: str
    expression_operator: str
    has_grouped_values: bool
    def __init__(self, *values: Any) -> None: ...
    def __and__(self, other: ConditionBase) -> And: ...
    def __or__(self, other: ConditionBase) -> Or: ...
    def __invert__(self) -> Not: ...
    def get_expression(self) -> ConditionBaseExpressionType: ...
    def __eq__(self, other: Any) -> bool: ...
    def __ne__(self, other: Any) -> bool: ...

class AttributeBase(object):
    def __init__(self, name: str) -> None: ...
    def __and__(self, value: Any) -> Any: ...
    def __or__(self, value: Any) -> Any: ...
    def __invert__(self) -> Any: ...
    def eq(self, value: Any) -> Equals: ...
    def lt(self, value: Any) -> LessThan: ...
    def lte(self, value: Any) -> LessThanEquals: ...
    def gt(self, value: Any) -> GreaterThan: ...
    def gte(self, value: Any) -> GreaterThanEquals: ...
    def begins_with(self, value: Any) -> BeginsWith: ...
    def between(self, low_value: Any, high_value: Any) -> Between: ...
    def __eq__(self, other: Any) -> bool: ...
    def __ne__(self, other: Any) -> bool: ...

class ConditionAttributeBase(ConditionBase, AttributeBase):
    def __init__(self, *values: Any) -> None: ...
    def __eq__(self, other: Any) -> bool: ...
    def __ne__(self, other: Any) -> bool: ...

class ComparisonCondition(ConditionBase):
    expression_format: Literal["{0} {operator} {1}"]

class Equals(ComparisonCondition):
    expression_operator: Literal["="]

class NotEquals(ComparisonCondition):
    expression_operator: Literal["<>"]

class LessThan(ComparisonCondition):
    expression_operator: Literal["<"]

class LessThanEquals(ComparisonCondition):
    expression_operator: Literal["<="]

class GreaterThan(ComparisonCondition):
    expression_operator: Literal[">"]

class GreaterThanEquals(ComparisonCondition):
    expression_operator: Literal[">="]

class In(ComparisonCondition):
    expression_operator: Literal["IN"]
    has_grouped_values: bool

class Between(ConditionBase):
    expression_operator: Literal["BETWEEN"]
    expression_format: Literal["{0} {operator} {1} AND {2}"]

class BeginsWith(ConditionBase):
    expression_operator: Literal["begins_with"]
    expression_format: Literal["{operator}({0}, {1})"]

class Contains(ConditionBase):
    expression_operator: Literal["contains"]
    expression_format: Literal["{operator}({0}, {1})"]

class Size(ConditionAttributeBase):
    expression_operator: Literal["size"]
    expression_format: Literal["{operator}({0})"]

class AttributeType(ConditionBase):
    expression_operator: Literal["attribute_type"]
    expression_format: Literal["{operator}({0}, {1})"]

class AttributeExists(ConditionBase):
    expression_operator: Literal["attribute_exists"]
    expression_format: Literal["{operator}({0})"]

class AttributeNotExists(ConditionBase):
    expression_operator: Literal["attribute_not_exists"]
    expression_format: Literal["{operator}({0})"]

class And(ConditionBase):
    expression_operator: Literal["AND"]
    expression_format: Literal["({0} {operator} {1})"]

class Or(ConditionBase):
    expression_operator: Literal["OR"]
    expression_format: Literal["({0} {operator} {1})"]

class Not(ConditionBase):
    expression_operator: Literal["NOT"]
    expression_format: Literal["({operator} {0})"]

class Key(AttributeBase): ...

class Attr(AttributeBase):
    def ne(self, value: Any) -> NotEquals: ...
    def is_in(self, value: Any) -> In: ...
    def exists(self) -> AttributeExists: ...
    def not_exists(self) -> AttributeNotExists: ...
    def contains(self, value: Any) -> Contains: ...
    def size(self) -> Size: ...
    def attribute_type(self, value: Any) -> AttributeType: ...

BuiltConditionExpression = NamedTuple(
    "BuiltConditionExpression",
    [
        ("condition_expression", str),
        ("attribute_name_placeholders", List[str]),
        ("attribute_value_placeholders", List[str]),
    ],
)

class ConditionExpressionBuilder:
    def __init__(self) -> None: ...
    def reset(self) -> None: ...
    def build_expression(
        self, condition: ConditionBase, is_key_condition: bool = False
    ) -> BuiltConditionExpression: ...



"""
Jinja2 renderer and black formatter.
"""
from pathlib import Path
from typing import Optional

import black

from mypy_boto3_builder.constants import TEMPLATES_PATH, LINE_LENGTH
from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.structures.package import Package
from mypy_boto3_builder.jinja_manager import JinjaManager


def blackify(content: str, file_path: Path, fast: bool = True) -> str:
    """
    Format `content` with `black` if `file_path` is `*.py` or `*.pyi`.

    On error writes invalid `content` to `file_path` to check for errors.

    Arguments:
        content -- Python code to format.
        file_path -- Target file path.
        fast -- Whether to skip AST post-check.

    Returns:
        Formatted python code.

    Raises:
        ValueError -- If `content` is not a valid Python code.
    """
    if file_path.suffix not in (".py", ".pyi"):
        return content

    file_mode = black.FileMode(
        is_pyi=file_path.suffix == ".pyi", line_length=LINE_LENGTH
    )
    try:
        content = black.format_file_contents(content, fast=fast, mode=file_mode)
    except black.NothingChanged:
        pass
    except IndentationError as e:
        file_path.write_text(content)
        raise ValueError(f"Cannot parse {file_path}: {e}")
    except black.InvalidInput as e:
        file_path.write_text(content)
        raise ValueError(f"Cannot parse {file_path}: {e}")

    return content


def render_jinja2_template(
    template_path: Path,
    package: Optional[Package] = None,
    service_name: Optional[ServiceName] = None,
) -> str:
    """
    Render Jinja2 template to a string.

    Arguments:
        template_path -- Relative path to template in `TEMPLATES_PATH`
        module -- Module record.
        service_name -- ServiceName instance.

    Returns:
        A rendered template.
    """
    template_full_path = TEMPLATES_PATH / template_path
    if not template_full_path.exists():
        raise ValueError(f"Template {template_path} not found")

    template = JinjaManager.get_environment().get_template(template_path.as_posix())
    return template.render(package=package, service_name=service_name)



"""
Parser that produces `structures.MasterModule`.
"""
from boto3.session import Session

from mypy_boto3_builder.structures.master_package import MasterPackage
from mypy_boto3_builder.parsers.fake_service_package import parse_fake_service_package


def parse_master_package(session: Session) -> MasterPackage:
    """
    Parse data for master package.

    Arguments:
        session -- boto3 session.

    Returns:
        MasterModule structure.
    """
    result = MasterPackage()
    for service_name in result.service_names:
        result.service_packages.append(
            parse_fake_service_package(session, service_name)
        )

    return result



"""
Helper for Python import strings.
"""
from __future__ import annotations

from typing import Any, Optional, Tuple
from functools import total_ordering

from mypy_boto3_builder.import_helpers.import_string import ImportString
from mypy_boto3_builder.constants import MODULE_NAME, TYPE_DEFS_NAME


@total_ordering
class ImportRecord:
    """
    Helper for Python import strings.

    Arguments:
        source -- Source of import.
        name -- Import name.
        alias -- Import local name.
        min_version -- Minimum Python version, used for fallback.
        fallback -- Fallback ImportRecord.
    """

    _is_internal = False
    type_defs_import_string = ImportString(TYPE_DEFS_NAME)
    builtins_import_string = ImportString("builtins")
    third_party_import_strings = (
        ImportString("boto3"),
        ImportString("botocore"),
    )

    def __init__(
        self,
        source: ImportString,
        name: str = "",
        alias: str = "",
        min_version: Tuple[int, ...] = (3, 8),
        fallback: Optional[ImportRecord] = None,
    ) -> None:
        self.source = source
        self.name = name
        self.alias = alias
        self.min_version = min_version
        self.fallback = fallback

    def __bool__(self) -> bool:
        return bool(self.source)

    @classmethod
    def empty(cls) -> ImportRecord:
        return cls(ImportString.empty())

    def render(self) -> str:
        if self.name and self.alias:
            return f"from {self.source} import {self.name} as {self.alias}"
        if self.name:
            return f"from {self.source} import {self.name}"
        if self.alias:
            return f"import {self.source} as {self.alias}"
        return f"import {self.source}"

    def __str__(self) -> str:
        return self.render()

    def __hash__(self) -> int:
        return hash(str(self))

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, ImportRecord):
            raise ValueError(f"Cannot compare ImportString with {other}")

        return str(self) == str(other)

    def __ne__(self, other: Any) -> bool:
        if not isinstance(other, ImportRecord):
            raise ValueError(f"Cannot compare ImportString with {other}")

        return not self == other

    def __gt__(self, other: ImportRecord) -> bool:
        if self.source == other.source:
            return self.name > other.name

        if self.fallback is not None and other.fallback is None:
            return True

        if other.fallback is not None and self.fallback is None:
            return False

        if self.is_local() and not other.is_local():
            return True

        if other.is_local() and not self.is_local():
            return False

        if self.is_third_party() and not other.is_third_party():
            return True

        if other.is_third_party() and not self.is_third_party():
            return False

        return self.source > other.source

    def get_local_name(self) -> str:
        """
        Get local import name.
        """
        return self.alias or self.name or self.source.render()

    def is_builtins(self) -> bool:
        """
        Whether import is from Python `builtins` module.
        """
        return self.source.startswith(self.builtins_import_string)

    def is_type_defs(self) -> bool:
        """
        Whether import is from `type_defs` module.
        """
        return self.source.startswith(self.type_defs_import_string)

    def is_third_party(self) -> bool:
        """
        Whether import is from 3rd party module.
        """
        for third_party_import_string in self.third_party_import_strings:
            if self.source.startswith(third_party_import_string):
                return True

        return False

    def is_local(self) -> bool:
        """
        Whether import is from local module.
        """
        if not self.source:
            return False

        if self.source.master_name.startswith(MODULE_NAME):
            return True

        if self.is_type_defs():
            return True

        return False

    def is_internal(self) -> bool:
        """
        Whether import is internal and requires `get_external` call before rendering.
        """
        return self._is_internal

    def get_external(self, _module_name: str) -> ImportRecord:
        """
        Get itself.

        Overriden by `InternalImportRecord`.
        """
        return self

    def is_standalone(self) -> bool:
        """
        Whether import record should not be grouped.
        """
        if not self.name or self.fallback:
            return True

        return False



# pylint: disable=unused-argument,multiple-statements,super-init-not-called,no-self-use,no-self-argument
from __future__ import annotations

import logging
from typing import Any, Iterator, List

from botocore.hooks import HierarchicalEmitter

from boto3.resources.model import Collection
from boto3.resources.base import ServiceResource
from boto3.resources.factory import ResourceFactory
from boto3.resources.response import ResourceHandler
from boto3.utils import ServiceContext

logger: logging.Logger

class ResourceCollection:
    def __init__(
        self,
        model: Collection,
        parent: ServiceResource,
        handler: ResourceHandler,
        **kwargs: Any
    ) -> None: ...
    def __repr__(self) -> str: ...
    def __iter__(self) -> Iterator[Any]: ...
    def pages(self) -> List[ServiceResource]: ...
    def all(self) -> ResourceCollection: ...
    def filter(self, **kwargs: Any) -> ResourceCollection: ...
    def limit(self, count: int) -> ResourceCollection: ...
    def page_size(self, count: int) -> ResourceCollection: ...

class CollectionManager:
    _collection_cls = ResourceCollection
    def __init__(
        self,
        collection_model: Collection,
        parent: ServiceResource,
        factory: ResourceFactory,
        service_context: ServiceContext,
    ) -> None: ...
    def __repr__(self) -> str: ...
    def iterator(self, **kwargs: Any) -> ResourceCollection: ...
    def all(self) -> ResourceCollection: ...
    def filter(self, **kwargs: Any) -> ResourceCollection: ...
    def limit(self, count: int) -> ResourceCollection: ...
    def page_size(self, count: InterruptedError) -> ResourceCollection: ...
    def pages(self) -> List[ServiceResource]: ...

class CollectionFactory:
    def load_from_definition(
        self,
        resource_name: str,
        collection_model: Collection,
        service_context: ServiceContext,
        event_emitter: HierarchicalEmitter,
    ) -> CollectionManager: ...



"""
CLI parser.
"""
import argparse
from pathlib import Path

from mypy_boto3_builder.service_name import ServiceName, ServiceNameCatalog
from mypy_boto3_builder.version import __version__ as version


def get_absolute_path(path: str) -> Path:
    """
    Get absolute path from a string.

    Arguments:
        path -- String containing path.

    Returns:
        Absolute path.
    """
    return Path(path).absolute()


def get_service_name(name: str) -> ServiceName:
    """
    Convert boto3 service name to ServiceName.

    Arguments:
        name -- Service name.

    Raises:
        argparse.ArgumentTypeError -- If service not found.
    """
    try:
        return ServiceNameCatalog.find(name)
    except ValueError as e:
        raise argparse.ArgumentTypeError(e)


def get_cli_parser() -> argparse.ArgumentParser:
    """
    Main CLI parser for builder.

    Returns:
        Argument parser.
    """
    parser = argparse.ArgumentParser(
        "mypy_boto3_builder", description="Builder for mypy-boto3."
    )
    parser.add_argument(
        "-d", "--debug", action="store_true", help="Show debug messages"
    )
    parser.add_argument(
        "-b",
        "--build-version",
        help="Set custom pacakges version, otherwise boto3 version is used.",
    )
    parser.add_argument("-v", "--version", action="version", version=version)
    parser.add_argument(
        "--skip-master",
        action="store_true",
        help="Whether to skip master and stubs modules",
    )
    parser.add_argument(
        "--skip-services", action="store_true", help="Whether to skip service modules"
    )
    parser.add_argument(
        "--panic",
        action="store_true",
        help="Raise exception on logger warning and above",
    )
    parser.add_argument(
        "output_path", metavar="OUTPUT_PATH", help="Output path", type=get_absolute_path
    )
    parser.add_argument(
        "-s",
        "--services",
        dest="service_names",
        nargs="*",
        metavar="SERVICE_NAME",
        help="List of AWS services, by default all services are used",
        type=get_service_name,
        default=ServiceNameCatalog.ITEMS,
    )
    return parser



"""
Structure for boto3-stubs module.
"""

from dataclasses import dataclass
from typing import List, Tuple

from mypy_boto3_builder.import_helpers.import_string import ImportString
from mypy_boto3_builder.service_name import ServiceName, ServiceNameCatalog
from mypy_boto3_builder.constants import BOTO3_STUBS_NAME, MODULE_NAME, TYPE_DEFS_NAME
from mypy_boto3_builder.structures.package import Package


@dataclass
class Boto3StubsPackage(Package):
    """
    Structure for boto3-stubs module.
    """

    name: str = BOTO3_STUBS_NAME
    pypi_name: str = BOTO3_STUBS_NAME

    MASTER_IMPORT_STRING = ImportString(MODULE_NAME)
    MASTER_TYPE_DEFS_IMPORT_STRING = ImportString(MODULE_NAME, TYPE_DEFS_NAME)

    @property
    def essential_service_names(self) -> List[ServiceName]:
        result: List[ServiceName] = []
        for service_name in self.service_names:
            if service_name.is_essential():
                result.append(service_name)
        return result

    @property
    def service_names(self) -> Tuple[ServiceName, ...]:
        return ServiceNameCatalog.ITEMS



"""
Wrapper for classes like `Paginator`.
"""
from __future__ import annotations

import inspect

from mypy_boto3_builder.import_helpers.import_string import ImportString
from mypy_boto3_builder.import_helpers.import_record import ImportRecord
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation


class TypeClass(FakeAnnotation):
    """
    Wrapper for classes like `Paginator`.

    Arguments:
        value -- Any Class.
        alias -- Local name.
    """

    def __init__(self, value: type, alias: str = "") -> None:
        self.value = value
        self.alias = alias

    def render(self, parent_name: str = "") -> str:
        """
        Render type annotation to a valid Python code for local usage.

        Returns:
            A string with a valid type annotation.
        """
        if self.alias:
            return self.alias

        return self.get_import_name()

    def get_import_name(self) -> str:
        return self.value.__name__

    def get_import_record(self) -> ImportRecord:
        module = inspect.getmodule(self.value)
        if module is None:
            raise ValueError(f"Unknown module for {self.value}")

        module_name = module.__name__
        return ImportRecord(
            source=ImportString.from_str(module_name),
            name=self.get_import_name(),
            alias=self.alias,
        )

    def copy(self) -> TypeClass:
        """
        Create a copy of type annotation wrapper.
        """
        return TypeClass(self.value, self.alias)

    def is_potential_literal(self) -> bool:
        """
        Whether type annotation can become a Literal.
        """
        if self.value is str:
            return True

        return False



"""
Multiple string utils collection.
"""
import re
import textwrap
from typing import Optional, List, Iterator

from mypy_boto3_builder.utils.indent_trimmer import IndentTrimmer


# Regexp to replace single backslashes
RE_BACKSLASH = re.compile(r"\\{1,2}")


def wrap_line(line: str, max_length: int) -> Iterator[str]:
    """
    Wrap text line to fit `max_length`.

    Arguments:
        line -- Text to wrap.
        max_length -- Result line max length.

    Yields:
        A string of wrapped text.
    """
    indent_length = len(line) - len(line.lstrip())
    indent = " " * indent_length
    max_line_length = max_length - indent_length
    for sub_line in textwrap.wrap(
        line.strip(), max_line_length, break_long_words=False, break_on_hyphens=False
    ):
        indented_sub_line = f"{indent}{sub_line}"
        if len(indented_sub_line) > max_line_length:
            yield from wrap_code_line(indented_sub_line, max_line_length)
            continue
        yield indented_sub_line


def wrap_code_line(line: str, max_length: int) -> Iterator[str]:
    """
    Wrap source code line to fit `max_length`.

    Arguments:
        line -- Source code text to wrap.
        max_length -- Result line max length.

    Yields:
        A string of wrapped text.
    """
    while len(line) > max_length:
        indent_length = len(line) - len(line.lstrip())
        indent = " " * indent_length

        equals_index = line.rfind("=", 0, max_length)
        if equals_index > 0:
            yield line[: equals_index + 1].rstrip()
            line = f"{indent}    {line[equals_index + 1 :]}"
            continue

        vertical_bar_index = line.rfind("|", indent_length + 1, max_length + 1)
        if vertical_bar_index > 0:
            yield line[:vertical_bar_index].rstrip()
            line = f"{indent}{line[vertical_bar_index :]}"
            continue

        vertical_bar_index = line.rfind("|", indent_length + 1)
        if vertical_bar_index > 0:
            yield line[:vertical_bar_index].rstrip()
            line = f"{indent}{line[vertical_bar_index :]}"
            continue

        yield line
        return

    yield line


def clean_doc(doc: Optional[str], max_length: int) -> str:
    """
    Clean docstring to be safely rendered.

    - If `doc` is not set, returns an empty docstring.
    - Returns extra empty lines.
    - Escapes backslashes.
    - Replace trible doublequotes with triple single quotes.
    - Tries to fit docstring to `max_length`

    Arguments:
        doc -- Instance docstring.
        max_length -- Result line max length.

    Returns:
        Cleaned docstring.
    """
    if doc is None:
        return ""

    lines = doc.split("\n")
    result: List[str] = []
    for line in lines:
        line = line.rstrip()
        line = RE_BACKSLASH.sub(r"\\\\", line)
        line = line.replace('"""', "'''")
        if not line and result and not result[-1]:
            continue

        if len(line) <= max_length:
            result.append(line)
            continue

        if " " in line.strip():
            for sub_line in wrap_line(line, max_length):
                result.append(sub_line)
            continue

        for sub_line in wrap_code_line(line, max_length):
            result.append(sub_line)

    while result and not result[-1].strip():
        result.pop()

    return "\n".join(result)


def get_class_prefix(func_name: str) -> str:
    """
    Get a valid Python class prefix from `func_name`.

    Arguments:
        func_name -- Any string.

    Returns:
        String with a class prefix.
    """
    parts = [f"{i[:1].upper()}{i[1:]}" for i in func_name.split("_") if i]
    return "".join(parts)


def get_line_with_indented(input_string: str, multi_first_line: bool = False) -> str:
    """
    Get first line of the string with all indented lines.

    Fixes invalid unindent.

    Arguments:
        input_string -- Input string.

    Returns:
        A string with first line and following indented lines.
    """
    result: List[str] = []
    indent_stack: List[int] = []
    for line in input_string.splitlines():
        line_indent = IndentTrimmer.get_line_indent(line)
        if not indent_stack:
            indent_stack.append(line_indent)
            result.append(line)
            continue

        if not line:
            result.append(line)
            continue

        if line_indent < indent_stack[0]:
            break

        if line_indent == indent_stack[0]:
            if not multi_first_line:
                break

            if len(indent_stack) > 1:
                break

        if line_indent == indent_stack[-1]:
            result.append(line)
            continue

        if line_indent > indent_stack[-1]:
            indent_stack.append(line_indent)
            result.append(line)
            continue

        while len(indent_stack) > 1 and line_indent <= indent_stack[-2]:
            indent_stack.pop()

        if indent_stack[-1] != line_indent:
            result.append(f"{' ' * indent_stack[-1]}{line[line_indent:]}")
            continue

        result.append(line)

    while result and not result[-1]:
        result.pop()

    return "\n".join(result)



"""
Main entrypoint for builder.
"""
from typing import List

from boto3 import __version__ as boto3_version
from boto3.session import Session

from mypy_boto3_builder.writers.processors import (
    process_boto3_stubs,
    process_master,
    process_service,
)
from mypy_boto3_builder.version import __version__ as version
from mypy_boto3_builder.logger import get_logger
from mypy_boto3_builder.utils.strings import clean_doc
from mypy_boto3_builder.cli_parser import get_cli_parser
from mypy_boto3_builder.service_name import ServiceName, ServiceNameCatalog
from mypy_boto3_builder.jinja_manager import JinjaManager
from mypy_boto3_builder.constants import (
    MODULE_NAME,
    DUMMY_REGION,
    BOTO3_STUBS_NAME,
    PYPI_NAME,
    LINE_LENGTH,
)


def main() -> None:
    """
    Main entrypoint for builder.
    """
    parser = get_cli_parser()
    args = parser.parse_args()
    logger = get_logger(verbose=args.debug, panic=args.panic)
    session = Session(region_name=DUMMY_REGION)
    args.output_path.mkdir(exist_ok=True)
    service_names: List[ServiceName] = []
    available_services = session.get_available_services()
    if len(args.service_names) == len(ServiceNameCatalog.ITEMS):
        for available_service in available_services:
            try:
                ServiceNameCatalog.find(available_service)
            except ValueError:
                logger.info(f"Service {available_service} is not supported, skipping.")

    for service_name in args.service_names:
        if service_name.name not in available_services:
            logger.warning(f"Service {service_name.name} is not available, skipping.")
            continue

        service_name.boto3_version = boto3_version
        service_names.append(service_name)

    build_version = args.build_version or boto3_version
    JinjaManager.update_globals(
        master_pypi_name=PYPI_NAME,
        master_module_name=MODULE_NAME,
        boto3_stubs_name=BOTO3_STUBS_NAME,
        boto3_version=boto3_version,
        build_version=build_version,
        doc_line_length=LINE_LENGTH - 4,
        clean_doc=clean_doc,
    )

    logger.info(f"Bulding version {build_version}")

    if not args.skip_services:
        for index, service_name in enumerate(service_names):
            logger.info(
                f"[{index + 1}/{len(service_names)}] Generating {service_name.module_name} module"
            )
            output_path = args.output_path / f"{service_name.module_name}_package"
            process_service(
                session=session, output_path=output_path, service_name=service_name
            )

    if not args.skip_master:
        logger.info(f"Generating {MODULE_NAME} module")
        output_path = args.output_path / "master_package"
        process_master(session, output_path)

        logger.info(f"Generating {BOTO3_STUBS_NAME} module")
        output_path = args.output_path / "boto3_stubs_package"
        process_boto3_stubs(output_path)

    logger.info(f"Completed")


if __name__ == "__main__":
    main()



"""
Helpers for parsing methods and attributes.
"""
import inspect
import textwrap
from typing import List, Dict, Any
from types import FunctionType

from boto3.resources.base import ServiceResource as Boto3ServiceResource

from mypy_boto3_builder.structures.method import Method
from mypy_boto3_builder.structures.attribute import Attribute
from mypy_boto3_builder.type_maps.docless_method_argument_map import (
    DOCLESS_METHOD_ARGUMENT_MAP,
)
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.utils.strings import get_class_prefix
from mypy_boto3_builder.parsers.docstring_parser.argspec_parser import ArgSpecParser
from mypy_boto3_builder.parsers.docstring_parser.docstring_parser import DocstringParser
from mypy_boto3_builder.logger import get_logger


def get_public_methods(inspect_class: Any) -> Dict[str, FunctionType]:
    """
    Extract public methods from any class.

    Arguments:
        inspect_class -- Inspect class.

    Returns:
        A dictionary of method name and method.
    """
    class_members = inspect.getmembers(inspect_class)
    methods: Dict[str, FunctionType] = {}
    for name, member in class_members:
        if not inspect.ismethod(member):
            continue

        if name.startswith("_"):
            continue

        methods[name] = member

    return methods


def parse_attributes(resource: Boto3ServiceResource) -> List[Attribute]:
    """
    Extract attributes from boto3 resource.

    Arguments:
        resource -- boto3 service resource.

    Returns:
        A list of Attribute structures.
    """
    result: List[Attribute] = []
    if not resource.meta.client:
        return result
    if not resource.meta.resource_model:
        return result

    service_model = resource.meta.client.meta.service_model
    if resource.meta.resource_model.shape:
        shape = service_model.shape_for(resource.meta.resource_model.shape)
        attributes = resource.meta.resource_model.get_attributes(shape)
        for name, attribute in attributes.items():
            result.append(
                Attribute(
                    name, DocstringParser.parse_type(attribute[1].type_name, name)
                )
            )

    return result


def parse_method(parent_name: str, name: str, method: FunctionType) -> Method:
    """
    Parse method to a structure.

    Arguments:
        parent_name -- Parent class name.
        method -- Inspect method.

    Returns:
        Method structure.
    """
    logger = get_logger()
    docstring = textwrap.dedent(inspect.getdoc(method) or "")
    return_type: FakeAnnotation = Type.none
    if not docstring:
        logger.info(f"Fixing {parent_name}.{name} method with no docstring")
        return Method(
            name=name,
            arguments=DOCLESS_METHOD_ARGUMENT_MAP[f"{parent_name}.{name}"],
            return_type=return_type,
        )

    logger.debug(f"Slow parsing of {parent_name}.{name}: {len(docstring)} chars")
    prefix = f"{get_class_prefix(parent_name)}{get_class_prefix(name)}"
    arg_spec_parser = ArgSpecParser(prefix)
    arguments = arg_spec_parser.get_function_arguments(method)
    docstring_parser = DocstringParser(prefix, arguments)
    arguments = docstring_parser.get_arguments(docstring)
    return_type = docstring_parser.get_return_type(docstring)

    return Method(name=name, arguments=arguments, return_type=return_type)



"""
Wrapper for `typing/typing_extensions.Literal` type annotations like `Literal['a', 'b']`
"""
from __future__ import annotations

from typing import Any

from mypy_boto3_builder.import_helpers.import_record import ImportRecord
from mypy_boto3_builder.import_helpers.import_string import ImportString
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation


class TypeLiteral(FakeAnnotation):
    """
    Wrapper for `typing/typing_extensions.Literal` type annotations like `Literal['a', 'b']`

    Arguments:
        children -- Literal values.
    """

    def __init__(self, *children: Any) -> None:
        self.children = list(children)

    def render(self, parent_name: str = "") -> str:
        """
        Render type annotation to a valid Python code for local usage.

        Returns:
            A string with a valid type annotation.
        """
        if not self.children:
            raise ValueError(f"Empty children for literal")
        children = ", ".join([repr(i) for i in self.children])
        return f"Literal[{children}]"

    @staticmethod
    def get_import_record() -> ImportRecord:
        """
        Get import record required for using type annotation.
        """
        return ImportRecord(
            ImportString("typing"),
            "Literal",
            min_version=(3, 8),
            fallback=ImportRecord(ImportString("typing_extensions"), "Literal"),
        )

    def copy(self) -> TypeLiteral:
        """
        Create a copy of type annotation wrapper.
        """
        return TypeLiteral(*self.children)

    def is_literal(self) -> bool:  # pylint: disable=no-self-use
        """
        Whether type annotation is `Literal`.
        """
        return True

    def add_child(self, child: FakeAnnotation) -> None:
        raise ValueError("Use add_literal_child function.")

    def add_literal_child(self, child: Any) -> None:
        """
        Add new child to `TypeLiteral` annotation.
        """
        self.children.append(child)



"""
Wrapper for type annotations imported from 3rd party libraries, like `boto3.service.Service`.
"""
from __future__ import annotations

from mypy_boto3_builder.import_helpers.import_record import ImportRecord
from mypy_boto3_builder.import_helpers.import_string import ImportString
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation


class ExternalImport(FakeAnnotation):
    """
    Wrapper for type annotations imported from 3rd party libraries, like
    `boto3.service.Service`.

    Arguments:
        source -- Module import string.
        name -- Import name.
        alias -- Import local name.
    """

    def __init__(self, source: ImportString, name: str = "", alias: str = "",) -> None:
        self.source = source
        self.name = name
        self.alias = alias
        self.import_record = ImportRecord(source=source, name=name, alias=alias)

    def render(self, parent_name: str = "") -> str:
        """
        Get string with local name to use.

        Returns:
            Import record local name.
        """
        return self.import_record.get_local_name()

    def get_import_record(self) -> ImportRecord:
        """
        Get import record required for using type annotation.
        """
        return self.import_record

    def copy(self) -> ExternalImport:
        """
        Create a copy of type annotation wrapper.
        """
        return ExternalImport(self.source, self.name, self.alias)



from pathlib import Path

from setuptools import setup, find_packages

from mypy_boto3_builder.version import __version__ as version


ROOT_PATH = Path(__file__).absolute().parent
README_PATH = ROOT_PATH.parent / "README.md"
REQUIREMENTS_PATH = ROOT_PATH / "requirements.txt"


LONG_DESCRIPTION = README_PATH.read_text()
INSTALL_REQUIRES = REQUIREMENTS_PATH.read_text().split("\n")


setup(
    name="mypy-boto3-builder",
    version=version,
    packages=find_packages(),
    url="https://github.com/vemel/mypy_boto3",
    license="MIT License",
    package_data={
        "mypy_boto3_builder": [
            "templates/*/*.jinja2",
            "templates/*/*/*.jinja2",
            "templates/*/*/*/*.jinja2",
            "boto3_stubs_static/*.pyi",
            "boto3_stubs_static/*/*.pyi",
            "boto3_stubs_static/*/*.py",
        ]
    },
    author="Vlad Emelianov",
    author_email="vlad.emelianov.nz@gmail.com",
    description="Builder for mypy-boto3.",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: Implementation :: CPython",
        "Typing :: Typed",
    ],
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    entry_points={
        "console_scripts": ["mypy_boto3_builder = mypy_boto3_builder.main:main"]
    },
    install_requires=INSTALL_REQUIRES,
    zip_safe=False,
)



"""
Utility for removing indentation for sections and lines.
"""

from typing import Iterable, List


__all__ = ["IndentTrimmer"]


class IndentTrimmer:
    """
    Utility for removing indentation for sections and lines.
    """

    @staticmethod
    def trim_empty_lines(lines: List[str]) -> List[str]:
        """
        Trim empty lines in the begging and the end of the text.

        Examples::

            IndentTrimmer.trim_empty_lines([
                '',
                '  asd',
                ' asd',
                '   asd',
                '  ',
                '',
            ])
            [
                '  asd',
                ' asd',
                '   asd',
            ]

        Returns:
            Lines wih no empty lines at start and end.
        """
        result = list(lines)
        while result and not result[0].strip():
            result.pop(0)
        while result and not result[-1].strip():
            result.pop()

        return result

    @classmethod
    def trim_text(cls, text: str) -> str:
        """
        Trim minimum indent from each line of text.

        Examples::

            IndentTrimmer.trim_text('  asd\\n asd\\n   asd\\n')
            ' asd\\nasd\\n  asd\\n'

        Arguments:
            text -- Multiline text.

        Returns:
            A text with trimmed indent.
        """
        new_lines = IndentTrimmer.trim_lines(text.split("\n"))
        return "\n".join(new_lines)

    @classmethod
    def trim_lines(cls, lines: Iterable[str]) -> List[str]:
        """
        Trim minimum indent from each line of text.

        Examples::

            IndentTrimmer.trim_lines([
                '  asd',
                ' asd',
                '   asd',
            ])
            [
                ' asd',
                'asd',
                '  asd',
            ]

        Arguments:
            lines -- List of lines.

        Returns:
            A list of lines with trimmed indent.
        """
        indents = [cls.get_line_indent(line) for line in lines if line.strip()]
        min_indent = 0
        if indents:
            min_indent = min(indents)

        new_lines = []
        for line in lines:
            new_lines.append(cls.trim_line(line, min_indent))

        return new_lines

    @staticmethod
    def trim_line(line: str, indent: int) -> str:
        """
        Trim indent from line if it is empty.

        Examples::

            IndentTrimmer.trim_line('     test', 2)
            '   test'

            IndentTrimmer.trim_line('     test', 6)
            'test'

            IndentTrimmer.trim_line('     test', 1)
            '    test'

        Arguments:
            line -- A line of text.

        Returns:
            A line with removed indent.
        """
        if not line[:indent].strip():
            return line[indent:]

        return line.lstrip()

    @staticmethod
    def get_line_indent(line: str) -> int:
        """
        Get indent length of the line.

        Examples::

            IndentTrimmer.get_line_indent('   test')
            3

            IndentTrimmer.get_line_indent('test')
            0

        Arguments:
            line -- Line of text.

        Returns:
            A number of indentation characters in a beginning of the line.
        """
        return len(line) - len(line.lstrip())

    @staticmethod
    def indent_line(line: str, indent: int) -> str:
        """
        Indent line with givent length `indent`

        Examples::

            IndentTrimmer.indent_line('test', 2)
            '  test'

        Arguments:
            line -- Line to indent.
            indent -- Length of indent in spaces.

        Returns:
            An indented line.
        """
        return u"{}{}".format(" " * indent, line)



# pylint: disable=unused-argument,multiple-statements,super-init-not-called,no-self-use,unused-import
from __future__ import annotations

import sys
import logging
from typing import Optional, Dict, Any, List, Union, Tuple

from botocore import xform_name
from botocore.model import Shape

if sys.version_info >= (3, 8):
    from typing import Literal, TypedDict
else:
    from typing_extensions import Literal, TypedDict

logger: logging.Logger

ActionDefinition = TypedDict(
    "ActionDefinition", {"request": Dict, "resource": Dict, "path": str}, total=False
)
DefinitionWithParamsDefinition = TypedDict(
    "DefinitionWithParamsDefinition", {"params": List[Dict]}, total=False
)
RequestDefinition = TypedDict("RequestDefinition", {"operation": str}, total=False)
WaiterDefinition = TypedDict("WaiterDefinition", {"waiterName": str}, total=False)
ResponseResourceDefinition = TypedDict(
    "ResponseResourceDefinition", {"type": str, "path": str}, total=False
)
ResourceModelDefinition = TypedDict(
    "ResourceModelDefinition", {"shape": str}, total=False
)

class Identifier:
    def __init__(self, name: str, member_name: Optional[str] = None) -> None:
        self.name: str
        self.member_name: str

class Action:
    def __init__(
        self, name: str, definition: ActionDefinition, resource_defs: Dict[str, Dict]
    ) -> None:
        self.name: str
        self.request: Optional[Request]
        self.resource: Optional[ResponseResource]
        self.path: Optional[str]

class DefinitionWithParams:
    def __init__(self, definition: DefinitionWithParamsDefinition) -> None: ...
    @property
    def params(self) -> List[Parameter]: ...

class Parameter:
    def __init__(
        self,
        target: str,
        source: str,
        name: str = None,
        path: str = None,
        value: Union[str, int, float, bool, None] = None,
        **kwargs: Any
    ) -> None:
        self.target: str
        self.source: str
        self.name: Optional[str]
        self.path: Optional[str]
        self.value: Union[str, int, float, bool, None]

class Request(DefinitionWithParams):
    def __init__(self, definition: RequestDefinition) -> None: ...

class Waiter(DefinitionWithParams):
    PREFIX: Literal["WaitUntil"]
    def __init__(self, name: str, definition: WaiterDefinition) -> None: ...

class ResponseResource:
    def __init__(
        self, definition: ResponseResourceDefinition, resource_defs: Dict[str, Dict]
    ) -> None:
        self.type: str
        self.path: str
    @property
    def identifiers(self) -> List[Identifier]: ...
    @property
    def model(self) -> ResourceModel: ...

class Collection(Action):
    @property
    def batch_actions(self) -> List[Action]: ...

class ResourceModel:
    def __init__(
        self,
        name: str,
        definition: ResourceModelDefinition,
        resource_defs: Dict[str, Dict],
    ) -> None:
        self.name: str
        self.shape: Optional[Shape]
    def load_rename_map(self, shape: Optional[Shape] = None) -> None: ...
    def get_attributes(self, shape: Shape) -> Dict[str, Tuple[str, Any]]: ...
    @property
    def identifiers(self) -> List[Identifier]: ...
    @property
    def load(self) -> Optional[Action]: ...
    @property
    def actions(self) -> List[Action]: ...
    @property
    def batch_actions(self) -> List[Action]: ...
    @property
    def subresources(self) -> List[ResponseResource]: ...
    @property
    def references(self) -> List[ResponseResource]: ...
    @property
    def collections(self) -> List[Collection]: ...
    @property
    def waiters(self) -> List[Waiter]: ...



"""
Class or module attribute.
"""
from dataclasses import dataclass
from typing import Set, Optional

from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.type_annotations.type_constant import TypeConstant


@dataclass
class Attribute:
    """
    Class or module attribute.

    Attributes:
        name -- Attribute name.
        type -- Attribute type annotation.
        value -- Attribute value.
    """

    name: str
    type: FakeAnnotation
    value: Optional[TypeConstant] = None

    def get_types(self) -> Set[FakeAnnotation]:
        """
        Return all type annotations used.

        Returns:
            A set of type annotations.
        """
        return self.type.get_types()



"""
Helper for Python import strings with not set master module name.
"""
from mypy_boto3_builder.enums.service_module_name import ServiceModuleName
from mypy_boto3_builder.import_helpers.import_record import ImportRecord
from mypy_boto3_builder.import_helpers.import_string import ImportString


class InternalImportRecord(ImportRecord):
    """
    Helper for Python import strings with not set master module name.

    Arguments:
        service_module_name -- Service module name.
        name -- Import name.
        alias -- Import local name.
    """

    _is_internal = True

    def __init__(
        self, service_module_name: ServiceModuleName, name: str = "", alias: str = ""
    ):
        super().__init__(ImportString(service_module_name.name), name=name, alias=alias)

    def get_external(self, module_name: str) -> ImportRecord:
        """
        Get full import record with `module_name` set as master module.

        Arguments:
            module_name -- Master module import string.

        Returns:
            A new non-internal ImportRecord.
        """
        source = ImportString(module_name) + self.source
        return ImportRecord(source=source, name=self.name, alias=self.alias,)



"""
String to type annotation map that find type annotation by argument name and type.
"""
from typing import Dict

from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.type_annotations.type_constant import TypeConstant
from mypy_boto3_builder.type_annotations.type_subscript import TypeSubscript
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_maps.typed_dicts import s3_copy_source_type, ec2_tag_type


ARGUMENT_TYPE_MAP: Dict[str, FakeAnnotation] = {
    "DryRun: bool": TypeConstant(False),
    "Tags: dict": TypeSubscript(Type.List, [ec2_tag_type]),
    "Tags: list": TypeSubscript(Type.List, [ec2_tag_type]),
    "CopySource: dict": s3_copy_source_type,
    "CopySource: str or dict": TypeSubscript(
        Type.Union, [s3_copy_source_type, Type.str]
    ),
}



from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_annotations.type_typed_dict import (
    TypeTypedDict,
    TypedDictAttribute,
)

s3_copy_source_type = TypeTypedDict(
    "CopySourceTypeDef",
    [
        TypedDictAttribute("Bucket", Type.str, True),
        TypedDictAttribute("Key", Type.str, True),
        TypedDictAttribute("VersionId", Type.str, False),
    ],
)


ec2_tag_type = TypeTypedDict(
    "TagTypeDef",
    [
        TypedDictAttribute("Key", Type.str, True),
        TypedDictAttribute("Value", Type.str, False),
    ],
)

waiter_config_type = TypeTypedDict(
    "WaiterConfigTypeDef",
    [
        TypedDictAttribute("Delay", Type.int, False),
        TypedDictAttribute("MaxAttempts", Type.int, False),
    ],
)

paginator_config_type = TypeTypedDict(
    "PaginatorConfigTypeDef",
    [
        TypedDictAttribute("MaxItems", Type.int, False),
        TypedDictAttribute("PageSize", Type.int, False),
        TypedDictAttribute("StartingToken", Type.str, False),
    ],
)



"""
Boto3 Client.
"""
from dataclasses import dataclass, field
from typing import List, Set

from botocore.client import BaseClient

from mypy_boto3_builder.service_name import ServiceName, ServiceNameCatalog
from mypy_boto3_builder.enums.service_module_name import ServiceModuleName
from mypy_boto3_builder.import_helpers.import_record import ImportRecord
from mypy_boto3_builder.import_helpers.import_string import ImportString
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.type_annotations.external_import import ExternalImport
from mypy_boto3_builder.structures.class_record import ClassRecord


@dataclass
class Client(ClassRecord):
    """
    Boto3 Client.
    """

    name: str = "Client"
    service_name: ServiceName = ServiceNameCatalog.ec2
    boto3_client: BaseClient = None
    exceptions_class: ClassRecord = field(
        default_factory=lambda: ClassRecord(name="Exceptions")
    )
    bases: List[FakeAnnotation] = field(
        default_factory=lambda: [
            ExternalImport(source=ImportString("botocore", "client"), name="BaseClient")
        ]
    )

    def __hash__(self) -> int:
        return hash(self.service_name)

    def get_import_records(self) -> Set[ImportRecord]:
        source = ImportString(
            self.service_name.module_name, ServiceModuleName.client.name
        )
        return {ImportRecord(source, self.name)}

    def get_all_names(self) -> List[str]:
        return [self.name]



"""
Wrapper for subscript type annotations, like `List[str]`.
"""
from __future__ import annotations

from typing import Set, Iterable

from mypy_boto3_builder.import_helpers.import_record import ImportRecord
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.type_annotations.type_annotation import TypeAnnotation


class TypeSubscript(FakeAnnotation):
    """
    Wrapper for subscript type annotations, like `List[str]`.

    Arguments:
        parent -- Parent type annotation.
        children -- Children type annotations.
    """

    def __init__(
        self, parent: TypeAnnotation, children: Iterable[FakeAnnotation] = (),
    ) -> None:
        self.parent = parent
        self.children = list(children)

    def __hash__(self) -> int:
        return hash(f"{self.parent}.{self.children}")

    def render(self, parent_name: str = "") -> str:
        """
        Render type annotation to a valid Python code for local usage.

        Returns:
            A string with a valid type annotation.
        """
        if not self.children:
            return f"{self.parent.render()}"

        children = ", ".join([i.render(parent_name) for i in self.children])
        return f"{self.parent.render()}[{children}]"

    def get_import_record(self) -> ImportRecord:
        """
        Get import record required for using type annotation.
        """
        return self.parent.get_import_record()

    def get_types(self) -> Set[FakeAnnotation]:
        result = self.parent.get_types()
        for child in self.children:
            result.update(child.get_types())
        return result

    def remove_children(self) -> None:
        self.children.clear()

    def add_child(self, child: FakeAnnotation) -> None:
        self.children.append(child)

    def is_dict(self) -> bool:
        return self.parent.is_dict()

    def is_list(self) -> bool:
        return self.parent.is_list()

    def copy(self) -> TypeSubscript:
        """
        Create a copy of type annotation wrapper.
        """
        return TypeSubscript(self.parent, list(self.children))



"""
Structure for parsed as dict `:type:` or `:rtype:` nested lines.
"""
from __future__ import annotations

from typing import Iterable, List, Any


class TypeDocLine:
    """
    Structure for parsed as dict `:type:` or `:rtype:` nested lines.

    Arguments:
        name -- Argument or TypedDict key name
        type_name -- Argument or TypedDict key type string.
        line -- Raw original line parts.
        description -- Rest of line for argument or TypedDict key definition.
        indented -- Intended lines.
    """

    def __init__(
        self,
        name: str = "",
        type_name: str = "",
        line: Iterable[str] = tuple(),
        description: str = "",
        indented: Iterable[Any] = tuple(),
    ) -> None:
        self.line = "".join(line)
        self.name = name
        self.type_name = type_name
        self.description = description
        self._indented = indented

    @property
    def indented(self) -> List[TypeDocLine]:
        """
        Get indented lines list.

        Returns:
            A list of `TypeDocLine`.
        """
        result: List[TypeDocLine] = []
        for line in self._indented:
            result.append(TypeDocLine(**line))
        return result

    @property
    def required(self) -> bool:
        """
        Whether the argument or TypedDict key is required.
        """
        return "REQUIRED" in self.description or "**must**" in self.description

    def render(self) -> str:
        """
        Get original string with indentation.

        Returns:
            A string as close as possible to original.
        """
        result: List[str] = []
        indent = "  " if self.line else ""
        if self.line:
            result.append(self.line)
            result.append("")
        for indented_line in self.indented:
            for indented_line_line in indented_line.render().splitlines():
                result.append(f"{indent}{indented_line_line}")
        return "\n".join(result)



"""
Jinja2 `Environment` manager.
"""
from typing import Any

import jinja2

from mypy_boto3_builder.constants import TEMPLATES_PATH
from mypy_boto3_builder.version import __version__ as version


__all__ = ["JinjaManager"]


class JinjaManager:
    """
    Jinja2 `Environment` manager.
    """

    _environment = jinja2.Environment(
        loader=jinja2.FileSystemLoader(TEMPLATES_PATH.as_posix()),
        undefined=jinja2.StrictUndefined,
    )

    @classmethod
    def update_globals(cls, **kwargs: Any) -> None:
        """
        Update global variables in `jinja2.Environment`.

        Arguments:
            kwargs -- Globals to set.
        """
        cls._environment.globals.update(kwargs)

    @classmethod
    def get_environment(cls) -> jinja2.Environment:
        """
        Get `jinja2.Environment`.
        """
        return cls._environment



"""
Getters for boto3 client and resource from session.
"""
from typing import Optional

from boto3.exceptions import ResourceNotExistsError
from boto3.session import Session
from boto3.resources.base import ServiceResource as Boto3ServiceResource
from botocore.client import BaseClient

from mypy_boto3_builder.service_name import ServiceName


def get_boto3_client(session: Session, service_name: ServiceName) -> BaseClient:
    """
    Get boto3 client from `session`.

    Arguments:
        session -- boto3 session.
        service_name -- ServiceName instance.

    Returns:
        Boto3 client.
    """
    return session.client(service_name.boto3_name)  # type: ignore


def get_boto3_resource(
    session: Session, service_name: ServiceName
) -> Optional[Boto3ServiceResource]:
    """
    Get boto3 resource from `session`.

    Arguments:
        session -- boto3 session.
        service_name -- ServiceName instance.

    Returns:
        Boto3 resource or None.
    """
    try:
        return session.resource(service_name.boto3_name)  # type: ignore
    except ResourceNotExistsError:
        return None
